@@ -7,6 +7,7 @@ import numpy as np
 
 from pinecone.protos import core_pb2
 from pinecone.utils import load_numpy
+from pinecone.utils.sentry import sentry_decorator as sentry
 from .api_controller import ControllerAPI
 from .api_router import RouterAPI
 from .constants import Config
@@ -91,6 +92,7 @@ class Cursor:
         self._create_reqeust_fn = config.create_request_fn
         self._parse_response_fn = config.parse_response_fn or self.identity
 
+    @sentry
     def take(self, size: int):
         """Processes a given number of items and returns the results.
 
@@ -99,10 +101,12 @@ class Cursor:
         result = list(self._fetch_unbuffered(size=size))
         return result or None
 
+    @sentry
     def collect(self):
         """Processes all of the items and returns the results."""
         return list(self._fetch_unbuffered())
 
+    @sentry
     def stream(self) -> Iterator:
         """Processes all of the items, returns the results one batch at a time."""
         return self._fetch_unbuffered()
@@ -207,6 +211,7 @@ class Connection:
         self.grpc = GRPCClient(host=host, port=port, api_key=api_key, service_name=service_name, conn_timeout=timeout,
                                timeout=response_timeout, traceroute=traceroute)
 
+    @sentry
     def upsert(
         self,
         items: Iterable[Tuple[str, np.ndarray]],
@@ -216,6 +221,8 @@ class Connection:
     ) -> Cursor:
         """Inserts or updates items.
 
+        An item is an `(id, vector)` tuple.
+
         Insert an item into the index, or update by item id if the item's id already exists.
 
         :param items: tuples of the form (id, numpy.ndarray). Id's length is limited to 64 characters.
@@ -247,6 +254,17 @@ class Connection:
         )
         return Cursor(config=config, items=items, batch_size=batch_size)
 
+    @sentry
+    def unary_upsert(self, item: Tuple[str, np.ndarray], namespace: str = None) -> Optional[IndexResult]:
+        """Inserts or updates an item.
+        
+        This is the unary version of :py:meth:`upsert`, which upserts one item at a time.
+        """
+        result = self.upsert(items=[item], namespace=namespace, cursor_config={"unary": True}).collect()
+        return result[0] if result else None
+
+
+    @sentry
     def delete(
         self, ids: Iterable[str], namespace: str = None, batch_size: int = None, cursor_config: dict = None
     ) -> Cursor:
@@ -279,6 +297,17 @@ class Connection:
         )
         return Cursor(config=config, items=ids, batch_size=batch_size)
 
+    @sentry
+    def unary_delete(self, id: str, namespace: str = None) -> Optional[DeleteResult]:
+        """Deletes an item by its id.
+        
+        This is the unary version of :py:meth:`delete`, which deletes one item at a time.
+        """
+        result = self.delete(ids=[id], namespace=namespace, cursor_config={"unary": True}).collect()
+        return result[0] if result else None
+
+
+    @sentry
     def fetch(
         self, ids: Iterable[str], namespace: str = None, batch_size: int = None, cursor_config: dict = None
     ) -> Cursor:
@@ -314,6 +343,17 @@ class Connection:
         )
         return Cursor(config=config, items=ids, batch_size=batch_size)
 
+    @sentry
+    def unary_fetch(self, id: str, namespace: str = None) -> Optional[FetchResult]:
+        """Fetches an item by its id.
+        
+        This is the unary version of :py:meth:`fetch`, which fetches one item at a time.
+        """
+        result = self.fetch(ids=[id], namespace=namespace, cursor_config={"unary": True}).collect()
+        return result[0] if result else None
+
+
+    @sentry
     def query(
         self,
         queries: Iterable[np.ndarray],
@@ -367,6 +407,21 @@ class Connection:
         )
         return Cursor(config=config, items=queries, batch_size=batch_size)
 
+    @sentry
+    def unary_query(
+        self, query: np.ndarray, namespace: str = None, top_k: int = 10, include_data: bool = False
+    ) -> Optional[QueryResult]:
+        """Sends a query.
+        
+        This is the unary version of :py:meth:`query`, which sends one query at a time.
+        """
+        result = self.query(
+            queries=[query], namespace=namespace, top_k=top_k, include_data=include_data, cursor_config={"unary": True}
+        ).collect()
+        return result[0] if result else None
+
+
+    @sentry
     def info(self, namespace: str = None) -> Optional[InfoResult]:
         """Returns information of the index."""
 
@@ -383,6 +438,7 @@ class Connection:
         result = Cursor(config=cursor_config, items=[1]).take(1)
         return result[0] if result else None
 
+    @sentry
     def close(self):
         """Closes the connection."""
 
@@ -394,6 +450,7 @@ class Connection:
         self.grpc.secure = is_secure
 
 
+@sentry
 def connect(service_name: str = None, router_name: str = None, timeout: int = 0,
             response_timeout: int = 20, traceroute: bool = False) -> Connection:
     """Connects to a live service/router and returns an instance of :class:`Connection`.
