@@ -4,10 +4,10 @@
 
 from pinecone.functions.index import Index
 from pinecone.protos import core_pb2
-from pinecone.utils import load_numpy, dump_numpy
+from pinecone.utils import load_numpy, dump_numpy, logger
 from pinecone.utils.constants import MAX_ID_LENGTH
 import numpy as np
-
+import json
 from typing import Iterable, Tuple
 from abc import abstractmethod
 
@@ -76,13 +76,19 @@ class VectorIndex(Index):
     @staticmethod
     def downcast_data(data: np.ndarray) -> np.ndarray:
         if not data.dtype == np.float32:
-            return data.astype('float32', casting='same_kind')
+            return data.astype('float32', casting='same_kind', copy=False)
         return data
 
     def validate_dimension(self, data: np.ndarray):
         if self.dimension() and data.shape[1] != self.dimension():
             raise ValueError(f"Data dimension must be consistent {data.shape[1]} != {self.dimension()}")
 
+    def export_metadata(self):
+        logger.opt(raw=True).info(json.dumps({"recordType": "vectorIndex.metadata",
+                                              "name": self.name,
+                                              "d": self.dimension(),
+                                              **self.config}) + "\n")
+
     @staticmethod
     def validate_ids(ids: Iterable[str]):
         if MAX_ID_LENGTH == 0:
