@@ -12,4 +12,4 @@ class AllSelector(SocketSelector):
 
     @classmethod
     def select_socket(cls, orig_msg: 'core_pb2.Request', shards: int) -> List[Optional['core_pb2.Request']]:
-        return [orig_msg] * shards
+        return [orig_msg]
