@@ -7,10 +7,10 @@ from pinecone.utils import replica_kube_hostname, replica_name
 from pinecone.protos import core_pb2
 from pinecone.network.zmq.servlet import ZMQServlet
 from pinecone.utils import open_or_create, get_hostname, wal
-from pinecone.utils.constants import MAX_CLIENTS, SEND_TIMEOUT, ZMQ_LOG_PORT, MAX_MSGS_PENDING
+from pinecone.utils.constants import MAX_CLIENTS, SEND_TIMEOUT, ZMQ_LOG_PORT, MAX_MSGS_PENDING, RECV_TIMEOUT
 from pinecone.network.zmq.spec import ServletSpec, SocketType, Socket
 
-from typing import Iterable, Tuple, List
+from typing import Iterable, Tuple, List, Union
 from loguru import logger
 
 import concurrent.futures
@@ -39,7 +39,8 @@ class WALServlet(ZMQServlet):
         self.executor = concurrent.futures.ThreadPoolExecutor()
 
         self.log_in = self.init_socket(self.log_in_socket)
-        self.log_outs = [self.init_socket(log_out_socket) for log_out_socket in self.log_out_sockets()]
+        self.log_outs = [self.init_socket(log_out_socket, disable_loadbalance=True)
+                         for log_out_socket in self.log_out_sockets()]
 
         self.in_buffer = asyncio.Queue()
         self.in_msgs = asyncio.Queue(maxsize=MAX_MSGS_PENDING)
@@ -53,7 +54,7 @@ class WALServlet(ZMQServlet):
 
         self.log_compaction_started = False
 
-        self.last = None
+        self.last = core_pb2.Request()
         self.size = self.load_size()
 
         offset_path = os.path.join(path, 'offsets')
@@ -100,12 +101,13 @@ class WALServlet(ZMQServlet):
             await self.handle_log_msg(msg)
 
     async def poll_sock(self, sock: zmq.Socket):
+        loop = asyncio.get_event_loop()
         while True:
             msg = await self.recv_msg(sock)
             if self.use_wal(msg):
                 await self.in_msgs.put(msg)
             else:
-                await self.handle_msg(msg)
+                loop.create_task(self.handle_msg(msg))
 
     def start_polling(self) -> List[asyncio.Task]:
         loop = asyncio.get_event_loop()
@@ -117,8 +119,10 @@ class WALServlet(ZMQServlet):
 
     async def handle_msg_wal(self, msg: core_pb2.Request):
         new_offset = await self.put(msg)
+        client_id = msg.client_id
+        client_offset = msg.client_offset
         await self.handle_msg(msg)
-        await self.set_client_offset(msg.client_id, msg.client_offset + 1)
+        await self.set_client_offset(client_id, client_offset + 1)
         await self.ack(new_offset)
 
     async def handle_log_msg(self, msg: 'core_pb2.LogEntry'):
@@ -127,9 +131,11 @@ class WALServlet(ZMQServlet):
                 if await self.put(msg.entry, offset=msg.offset):
                     await self.handle_msg(msg.entry)
             except InvalidOffsetException:
+                logger.warning(f"Received invalid offset {msg.offset}")
                 await self.send_log_msg(self.get_replay_request(), 0)
         else:
             await self.remote_ack(msg.offset, msg.ack.replica, msg.ack.replay)
+            del msg
 
     async def consume_msgs(self):
         while True:
@@ -154,14 +160,8 @@ class WALServlet(ZMQServlet):
         msg_pb.ParseFromString(msg)
         return msg_pb
 
-    async def send_log_msg(self, msg: 'core_pb2.LogEntry', replica: int):
-        assert replica != self.spec.replica
-        while True:
-            try:
-                await asyncio.wait_for(self.log_outs[replica].send(msg.SerializeToString()), SEND_TIMEOUT)
-                break
-            except asyncio.TimeoutError:
-                logger.error("SENDING TIMED OUT, RETRYING")
+    async def send_log_msg(self, msg: Union['core_pb2.LogEntry', bytes], replica: int):
+        await self.log_outs[replica].send(msg if type(msg) == bytes else msg.SerializeToString())
 
     async def run_sync(self, func, *args, **kwargs):
         loop = asyncio.get_event_loop()
@@ -217,7 +217,6 @@ class WALServlet(ZMQServlet):
                 wal.write_log_entry(self.log_file, entry)
                 self.size += 1
                 # we need to copy the message, because it might get changed later on
-                self.last = core_pb2.Request()
                 self.last.CopyFrom(entry)
                 logfile_size = self.log_file.tell()
             return logfile_size, self.size
@@ -247,14 +246,20 @@ class WALServlet(ZMQServlet):
             self.log_file.seek(0)
             pos = 0
             while True:
-                entry, msg_bytes = wal.read_entry(self.log_file)
+                entry, msg_size = wal.read_entry(self.log_file)
 
-                if msg_bytes == b'' or (end_offset and pos >= end_offset):
+                if msg_size == 0 or (end_offset and pos >= end_offset):
                     break
                 if pos >= offset:
                     yield pos, entry
                 pos += 1
 
+    def handle_replay_request(self, loop: asyncio.AbstractEventLoop, replica: int, offset: int, end_offset: int = None):
+        logger.info(f"Replaying log from offset {offset} to replica {replica}")
+        for pos, entry in self.sync_replay_log(offset, end_offset=end_offset):
+            asyncio.run_coroutine_threadsafe(self.send_log_msg(core_pb2.LogEntry(entry=entry,
+                                                                                 offset=pos), replica), loop)
+
     def get_leader_log_backlog(self, replica: int) -> Iterable[core_pb2.LogEntry]:
         return [core_pb2.LogEntry(entry=entry, offset=offset)
                 for offset, entry in self.sync_replay_log(self.get_offset(replica))]
@@ -270,11 +275,12 @@ class WALServlet(ZMQServlet):
         #     logger.warning('Acked by more than one offset')
 
         await self.set_offset(self.replica, offset)
-
         if self.leader:
+            msg = core_pb2.LogEntry(entry=self.last, offset=offset)
+            serialized_msg = msg.SerializeToString()
             for r in range(1, self.num_replicas):
-                msg = core_pb2.LogEntry(entry=self.last, offset=offset)
-                await self.send_log_msg(msg, r)
+                await self.send_log_msg(serialized_msg, r)
+
         else:
             ack = core_pb2.Ack(replica=self.replica)
             await self.send_log_msg(core_pb2.LogEntry(offset=offset, ack=ack), 0)
@@ -282,23 +288,23 @@ class WALServlet(ZMQServlet):
     async def remote_ack(self, start_offset: int, replica: int, replay=False):
         if replay:
             leader_offset = self.get_offset(self.replica)
-            logs = await self.run_sync(lambda s: list(self.sync_replay_log(s, end_offset=leader_offset)), start_offset)
-            if not leader_offset == len(logs) + start_offset:
-                logger.error(f"INCORRECT LENGTH OF REPLAY LOG {len(logs)} OFFSETS: {self.offsets_list()}. Size {self.size} start offset: {start_offset} ledaer: {leader_offset}")
-            for offset, msg in logs:
-                await self.send_log_msg(core_pb2.LogEntry(entry=msg, offset=offset), replica)
+            event_loop = asyncio.get_event_loop()
+            await self.run_sync(self.handle_replay_request, event_loop, replica, start_offset, end_offset=leader_offset)
 
-    def get_replay_request(self):
+    def get_replay_request(self, override_offset=None):
         ack = core_pb2.Ack(replica=self.replica, replay=True)
-        return core_pb2.LogEntry(offset=self.get_offset(self.replica), ack=ack)
+        offset = override_offset or self.get_offset(self.replica)
+        logger.info(f"Sending replay request for offset {offset}")
+        return core_pb2.LogEntry(offset=offset, ack=ack)
 
     def load_sync(self, handler):
         logger.info(f'Recovering from local WAL: offset: {self.get_offset(self.replica)}')
         for offset, entry in self.sync_replay_log(self.get_offset(self.replica)):
-            response = handler(entry)  # type: core_pb2.Request
-            self.sync_ack(offset)
-            if self.leader:
-                yield response
+            try:
+                handler(entry)  # type: core_pb2.Request
+                self.sync_ack(offset)
+            except Exception as e:
+                logger.error(e)
 
     def swap_log_files(self, end_file_offset: int):
         tmp_log_file = open(self.tmp_log_path, "r+b")
@@ -306,8 +312,8 @@ class WALServlet(ZMQServlet):
         with self.log_lock:
             self.log_file.seek(end_file_offset)
             while True:
-                entry, msg_bytes = wal.read_entry(self.log_file)
-                if msg_bytes == b'':
+                entry, msg_size = wal.read_entry(self.log_file)
+                if msg_size == 0:
                     break
                 wal.write_log_entry(tmp_log_file, entry)
             tmp_log_file.close()
